# This file is needed to package build_template.sh.

# Versions should comply with PEP440. For a discussion on single-sourcing the
# version across setup.py and the project code, see
# https://packaging.python.org/en/latest/single_source_version.html
#
# v1.9.0.rc1 is actually a placeholder which will be automatically replaced by
# the release-alibuild pipeline in jenkins whenever we need a new release.
__version__ = 'v1.9.0.rc1'
